from agents.agent import Agent


instruction_prompt = """You are a helpful chatbot with the role of a history tutor. 

Your purpose is to help students with their history homework and to prepare for exams.

"""

# history_tutor_agent = Agent(instruction_prompt)

# while True:
#     response_msg = history_tutor_agent.chat(input("Input message: "))
#     print(response_msg)



import importlib.util
import os
from typing import List, Callable

import maihem.errors as errors
from maihem.logger import get_logger

logger = get_logger()


def spread_n_into_buckets(n: int, buckets: int) -> List[int]:
    if not isinstance(n, int) or not isinstance(buckets, int):
        raise TypeError("n and buckets must be integers")
    if n < 1 or buckets < 1:
        raise ValueError("n and buckets must be greater than or equal to 1")

    return [
        max(1, n // buckets + (1 if x < n % buckets else 0)) for x in range(buckets)
    ]


def import_wrapper_function(path: str = "wrapper_function.py") -> Callable:
    if not os.path.exists(path):
        errors.raise_not_found_error(f"Wrapper function file not found at: {path}")

    if not os.path.isfile(path):
        errors.raise_not_found_error(f"Path exists but is not a file: {path}")

    try:
        spec = importlib.util.spec_from_file_location(
            "wrapper_function", os.path.abspath(path)
        )
        if spec is None or spec.loader is None:
            errors.raise_not_found_error("Failed to load module specification")

        wrapper_module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(wrapper_module)

        if not hasattr(wrapper_module, "wrapper_function"):
            errors.raise_not_found_error("Module does not contain 'wrapper_function'")

        wrapper_function = wrapper_module.wrapper_function
        return wrapper_function
    except Exception as e:
        errors.raise_not_found_error(
            f"Wrapper function could not be imported. Error: {e}"
        )


def create_project_folder(name: str) -> None:
    file_content = """import requests
from typing import List, Tuple

def wrapper_function(
    conversation_id: str,
    maihem_agent_message: str | None,
    conversation_history: dict,
) -> Tuple[str, List[str]]:
    \"\"\"Callable function to wrap your target agent to be tested.\"\"\"
    # Call demo Maihem target agent for quickstart
    url = "https://demo-agent.maihem.ai/chat"
    payload = {"message": maihem_agent_message if maihem_agent_message else " "}
    response = requests.request("POST", url, json=payload).json()
    return response["message"], response["contexts"]
"""
    if not name or not isinstance(name, str):
        raise ValueError("Project name must be a non-empty string")

    # Sanitize the folder name to prevent directory traversal
    name = os.path.basename(name)

    try:
        os.makedirs(name, exist_ok=True, mode=0o755)  # Explicit permissions

        file_path = os.path.join(name, "wrapper_function.py")

        # Check if file already exists to prevent accidental overwrites
        if os.path.exists(file_path):
            logger.warning(f"File already exists at {file_path}")
            return

        with open(file_path, "w", encoding="utf-8") as file:
            file.write(file_content)

        logger.info(f"Project folder created: '{name}'")
    except OSError as e:
        logger.error(f"Failed to create project folder: {e}")
        raise



import click
from maihem.clients import Maihem
from maihem.utils.utils import import_wrapper_function, create_project_folder
from maihem.errors import RequestValidationError


# Create the main CLI group
@click.group()
@click.option(
    "--env",
    type=click.Choice(["local", "staging", "production"]),
    default="production",
    help="Environment to use (default: production)",
)
@click.pass_context
def cli(ctx, env):
    """Maihem CLI"""
    try:
        ctx.obj = Maihem(env=env)
    except RequestValidationError as e:
        click.echo(e)
        ctx.exit(1)


# Target agent commands
@cli.group()
def target_agent():
    """Manage target agents"""
    pass


@target_agent.command()
@click.option("--name", required=True, help="Unique name of the target agent")
@click.option("--label", help="Label for the target agent (optional)")
@click.option("--role", required=True, help="Role of the target agent")
@click.option("--description", required=True, help="Description of the target agent")
@click.option("--language", default="en", help="Language (default: en). Follow ISO 639")
@click.pass_obj
def create(maihem_client, name, label, role, description, language):
    """Create a target agent"""
    maihem_client.create_target_agent(
        name=name, label=label, role=role, description=description, language=language
    )
    create_project_folder(name)


# Test commands
@cli.group()
def test():
    """Manage tests"""
    pass


@test.command()
@click.option("--name", required=True, help="Name of the test")
@click.option("--label", help="Label for the test (optional)")
@click.option("--target-agent-name", required=True, help="Name of the target agent")
@click.option(
    "--initiating-agent",
    type=click.Choice(["maihem", "target"]),
    default="maihem",
    help="Initiating agent (maihem or target)",
)
@click.option("--modules", required=True, help="Modules to use")
@click.option(
    "--documents-path", required=False, help="Path to the folder with documents"
)
@click.option(
    "--number-conversations", type=int, required=True, help="Number of conversations"
)
@click.option(
    "--conversation-turns-max",
    type=int,
    default=10,
    help="Max turns per conversation (default: 10)",
)
@click.option(
    "--maihem-behavior-prompt",
    help="Prompt for the maihem agent behavior",
)
@click.option(
    "--maihem-goal-prompt",
    help="Describes the goal of the simulated personas",
)
@click.option(
    "--maihem-population-prompt",
    help="Describes the desired population of simulated personas",
)
@click.pass_obj
def create(
    maihem_client,
    name,
    label,
    target_agent_name,
    initiating_agent,
    modules,
    documents_path,
    number_conversations,
    conversation_turns_max,
    maihem_behavior_prompt,
    maihem_goal_prompt,
    maihem_population_prompt,
):
    """Create a test"""
    module_list = [m.strip() for m in modules.split(",")]
    maihem_client.create_test(
        name=name,
        label=label,
        modules=module_list,
        target_agent_name=target_agent_name,
        initiating_agent=initiating_agent,
        documents_path=documents_path,
        conversation_turns_max=conversation_turns_max,
        number_conversations=number_conversations,
        maihem_behavior_prompt=maihem_behavior_prompt,
        maihem_goal_prompt=maihem_goal_prompt,
        maihem_population_prompt=maihem_population_prompt,
    )


@test.command()
@click.option("--name", required=True, help="Name of the test run")
@click.option("--label", help="Label for the test run (optional)")
@click.option("--test-name", required=True, help="Name of the test")
@click.option(
    "--wrapper-function-path",
    default="wrapper_function.py",
    help="Path to the wrapper function",
)
@click.option(
    "--concurrent-conversations",
    type=int,
    default=10,
    help="Number of concurrent conversations",
)
@click.pass_obj
def run(
    maihem_client,
    name,
    label,
    test_name,
    wrapper_function_path,
    concurrent_conversations,
):
    """Run a test"""
    wrapper_function = import_wrapper_function(wrapper_function_path)
    maihem_client.run_test(
        name=name,
        label=label,
        test_name=test_name,
        wrapper_function=wrapper_function,
        concurrent_conversations=concurrent_conversations,
    )


# Test run commands
@cli.group()
def test_run():
    """Manage test runs"""
    pass


@test_run.command()
@click.option("--test-name", required=True, help="Name of the test")
@click.option("--test-run-name", required=True, help="Name of the test run")
@click.pass_obj
def get(maihem_client, test_name, test_run_name):
    """Get test run results"""
    test_run_results = maihem_client.get_test_run_result(
        test_name=test_name, test_run_name=test_run_name
    )
    click.echo(test_run_results)


if __name__ == "__main__":
    cli()



from typing import List, Dict, Any

import maihem.errors as errors
from maihem.utils.utils import spread_n_into_buckets


def map_module_list_to_metrics(
    modules: List[str], number_conversations: int
) -> Dict[str, Dict[str, Any]]:
    """Map modules to metrics."""
    metrics_config = {}
    for module in modules:
        if module == "cx":
            if 0 < number_conversations <= 4:
                metrics_config["qa_cx_goal_completion"] = number_conversations
            else:
                convs_list = spread_n_into_buckets(number_conversations, 4)
                metrics_config["qa_cx_goal_completion"] = convs_list[0]
                metrics_config["qa_cx_helpfulness"] = convs_list[1]
                metrics_config["qa_cx_retention"] = convs_list[2]
                metrics_config["qa_cx_nps"] = convs_list[3]
        elif module == "rag":
            convs_list = spread_n_into_buckets(number_conversations, 3)
            metrics_config["qa_rag_hallucination"] = convs_list[0]
            metrics_config["qa_rag_answer_relevance"] = convs_list[1]
            metrics_config["qa_rag_context_relevance"] = convs_list[2]
        else:
            raise errors.raise_not_found_error(
                f"Module '{module}' is not supported. Please check the supported modules in https://docs.maihem.ai/reference/metric-collection."
            )

    return metrics_config



import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.test_result_enum import TestResultEnum
from ..models.test_status_enum import TestStatusEnum
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.links import Links
    from ..models.test_run_conversation_scores import TestRunConversationScores
    from ..models.test_run_results_metric_scores_type_0 import TestRunResultsMetricScoresType0
    from ..models.test_run_results_module_group_scores_type_0 import TestRunResultsModuleGroupScoresType0
    from ..models.test_run_results_module_scores_type_0 import TestRunResultsModuleScoresType0


T = TypeVar("T", bound="TestRunResults")


@_attrs_define
class TestRunResults:
    """
    Attributes:
        id (str):
        created_at (datetime.datetime):
        updated_at (datetime.datetime):
        name (str):
        test_id (str):
        test_name (str):
        agent_target_id (str):
        agent_target_name (str):
        status (TestStatusEnum):
        label (Union[None, Unset, str]):
        test_label (Union[None, Unset, str]):
        agent_target_label (Union[None, Unset, str]):
        started_at (Union[None, Unset, datetime.datetime]):
        completed_at (Union[None, Unset, datetime.datetime]):
        result (Union[None, TestResultEnum, Unset]):
        result_score (Union[None, Unset, float]):
        result_score_change (Union[None, Unset, float]):
        links (Union['Links', None, Unset]):
        conversation_scores (Union['TestRunConversationScores', None, Unset]):
        metric_scores (Union['TestRunResultsMetricScoresType0', None, Unset]):
        module_scores (Union['TestRunResultsModuleScoresType0', None, Unset]):
        module_group_scores (Union['TestRunResultsModuleGroupScoresType0', None, Unset]):
    """

    id: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    name: str
    test_id: str
    test_name: str
    agent_target_id: str
    agent_target_name: str
    status: TestStatusEnum
    label: Union[None, Unset, str] = UNSET
    test_label: Union[None, Unset, str] = UNSET
    agent_target_label: Union[None, Unset, str] = UNSET
    started_at: Union[None, Unset, datetime.datetime] = UNSET
    completed_at: Union[None, Unset, datetime.datetime] = UNSET
    result: Union[None, TestResultEnum, Unset] = UNSET
    result_score: Union[None, Unset, float] = UNSET
    result_score_change: Union[None, Unset, float] = UNSET
    links: Union["Links", None, Unset] = UNSET
    conversation_scores: Union["TestRunConversationScores", None, Unset] = UNSET
    metric_scores: Union["TestRunResultsMetricScoresType0", None, Unset] = UNSET
    module_scores: Union["TestRunResultsModuleScoresType0", None, Unset] = UNSET
    module_group_scores: Union["TestRunResultsModuleGroupScoresType0", None, Unset] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.links import Links
        from ..models.test_run_conversation_scores import TestRunConversationScores
        from ..models.test_run_results_metric_scores_type_0 import TestRunResultsMetricScoresType0
        from ..models.test_run_results_module_group_scores_type_0 import TestRunResultsModuleGroupScoresType0
        from ..models.test_run_results_module_scores_type_0 import TestRunResultsModuleScoresType0

        id = self.id

        created_at = self.created_at.isoformat()

        updated_at = self.updated_at.isoformat()

        name = self.name

        test_id = self.test_id

        test_name = self.test_name

        agent_target_id = self.agent_target_id

        agent_target_name = self.agent_target_name

        status = self.status.value

        label: Union[None, Unset, str]
        if isinstance(self.label, Unset):
            label = UNSET
        else:
            label = self.label

        test_label: Union[None, Unset, str]
        if isinstance(self.test_label, Unset):
            test_label = UNSET
        else:
            test_label = self.test_label

        agent_target_label: Union[None, Unset, str]
        if isinstance(self.agent_target_label, Unset):
            agent_target_label = UNSET
        else:
            agent_target_label = self.agent_target_label

        started_at: Union[None, Unset, str]
        if isinstance(self.started_at, Unset):
            started_at = UNSET
        elif isinstance(self.started_at, datetime.datetime):
            started_at = self.started_at.isoformat()
        else:
            started_at = self.started_at

        completed_at: Union[None, Unset, str]
        if isinstance(self.completed_at, Unset):
            completed_at = UNSET
        elif isinstance(self.completed_at, datetime.datetime):
            completed_at = self.completed_at.isoformat()
        else:
            completed_at = self.completed_at

        result: Union[None, Unset, str]
        if isinstance(self.result, Unset):
            result = UNSET
        elif isinstance(self.result, TestResultEnum):
            result = self.result.value
        else:
            result = self.result

        result_score: Union[None, Unset, float]
        if isinstance(self.result_score, Unset):
            result_score = UNSET
        else:
            result_score = self.result_score

        result_score_change: Union[None, Unset, float]
        if isinstance(self.result_score_change, Unset):
            result_score_change = UNSET
        else:
            result_score_change = self.result_score_change

        links: Union[Dict[str, Any], None, Unset]
        if isinstance(self.links, Unset):
            links = UNSET
        elif isinstance(self.links, Links):
            links = self.links.to_dict()
        else:
            links = self.links

        conversation_scores: Union[Dict[str, Any], None, Unset]
        if isinstance(self.conversation_scores, Unset):
            conversation_scores = UNSET
        elif isinstance(self.conversation_scores, TestRunConversationScores):
            conversation_scores = self.conversation_scores.to_dict()
        else:
            conversation_scores = self.conversation_scores

        metric_scores: Union[Dict[str, Any], None, Unset]
        if isinstance(self.metric_scores, Unset):
            metric_scores = UNSET
        elif isinstance(self.metric_scores, TestRunResultsMetricScoresType0):
            metric_scores = self.metric_scores.to_dict()
        else:
            metric_scores = self.metric_scores

        module_scores: Union[Dict[str, Any], None, Unset]
        if isinstance(self.module_scores, Unset):
            module_scores = UNSET
        elif isinstance(self.module_scores, TestRunResultsModuleScoresType0):
            module_scores = self.module_scores.to_dict()
        else:
            module_scores = self.module_scores

        module_group_scores: Union[Dict[str, Any], None, Unset]
        if isinstance(self.module_group_scores, Unset):
            module_group_scores = UNSET
        elif isinstance(self.module_group_scores, TestRunResultsModuleGroupScoresType0):
            module_group_scores = self.module_group_scores.to_dict()
        else:
            module_group_scores = self.module_group_scores

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "created_at": created_at,
                "updated_at": updated_at,
                "name": name,
                "test_id": test_id,
                "test_name": test_name,
                "agent_target_id": agent_target_id,
                "agent_target_name": agent_target_name,
                "status": status,
            }
        )
        if label is not UNSET:
            field_dict["label"] = label
        if test_label is not UNSET:
            field_dict["test_label"] = test_label
        if agent_target_label is not UNSET:
            field_dict["agent_target_label"] = agent_target_label
        if started_at is not UNSET:
            field_dict["started_at"] = started_at
        if completed_at is not UNSET:
            field_dict["completed_at"] = completed_at
        if result is not UNSET:
            field_dict["result"] = result
        if result_score is not UNSET:
            field_dict["result_score"] = result_score
        if result_score_change is not UNSET:
            field_dict["result_score_change"] = result_score_change
        if links is not UNSET:
            field_dict["links"] = links
        if conversation_scores is not UNSET:
            field_dict["conversation_scores"] = conversation_scores
        if metric_scores is not UNSET:
            field_dict["metric_scores"] = metric_scores
        if module_scores is not UNSET:
            field_dict["module_scores"] = module_scores
        if module_group_scores is not UNSET:
            field_dict["module_group_scores"] = module_group_scores

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.links import Links
        from ..models.test_run_conversation_scores import TestRunConversationScores
        from ..models.test_run_results_metric_scores_type_0 import TestRunResultsMetricScoresType0
        from ..models.test_run_results_module_group_scores_type_0 import TestRunResultsModuleGroupScoresType0
        from ..models.test_run_results_module_scores_type_0 import TestRunResultsModuleScoresType0

        d = src_dict.copy()
        id = d.pop("id")

        created_at = isoparse(d.pop("created_at"))

        updated_at = isoparse(d.pop("updated_at"))

        name = d.pop("name")

        test_id = d.pop("test_id")

        test_name = d.pop("test_name")

        agent_target_id = d.pop("agent_target_id")

        agent_target_name = d.pop("agent_target_name")

        status = TestStatusEnum(d.pop("status"))

        def _parse_label(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        label = _parse_label(d.pop("label", UNSET))

        def _parse_test_label(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        test_label = _parse_test_label(d.pop("test_label", UNSET))

        def _parse_agent_target_label(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        agent_target_label = _parse_agent_target_label(d.pop("agent_target_label", UNSET))

        def _parse_started_at(data: object) -> Union[None, Unset, datetime.datetime]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                started_at_type_0 = isoparse(data)

                return started_at_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, datetime.datetime], data)

        started_at = _parse_started_at(d.pop("started_at", UNSET))

        def _parse_completed_at(data: object) -> Union[None, Unset, datetime.datetime]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                completed_at_type_0 = isoparse(data)

                return completed_at_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, datetime.datetime], data)

        completed_at = _parse_completed_at(d.pop("completed_at", UNSET))

        def _parse_result(data: object) -> Union[None, TestResultEnum, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                result_type_0 = TestResultEnum(data)

                return result_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, TestResultEnum, Unset], data)

        result = _parse_result(d.pop("result", UNSET))

        def _parse_result_score(data: object) -> Union[None, Unset, float]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, float], data)

        result_score = _parse_result_score(d.pop("result_score", UNSET))

        def _parse_result_score_change(data: object) -> Union[None, Unset, float]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, float], data)

        result_score_change = _parse_result_score_change(d.pop("result_score_change", UNSET))

        def _parse_links(data: object) -> Union["Links", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                links_type_0 = Links.from_dict(data)

                return links_type_0
            except:  # noqa: E722
                pass
            return cast(Union["Links", None, Unset], data)

        links = _parse_links(d.pop("links", UNSET))

        def _parse_conversation_scores(data: object) -> Union["TestRunConversationScores", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                conversation_scores_type_0 = TestRunConversationScores.from_dict(data)

                return conversation_scores_type_0
            except:  # noqa: E722
                pass
            return cast(Union["TestRunConversationScores", None, Unset], data)

        conversation_scores = _parse_conversation_scores(d.pop("conversation_scores", UNSET))

        def _parse_metric_scores(data: object) -> Union["TestRunResultsMetricScoresType0", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                metric_scores_type_0 = TestRunResultsMetricScoresType0.from_dict(data)

                return metric_scores_type_0
            except:  # noqa: E722
                pass
            return cast(Union["TestRunResultsMetricScoresType0", None, Unset], data)

        metric_scores = _parse_metric_scores(d.pop("metric_scores", UNSET))

        def _parse_module_scores(data: object) -> Union["TestRunResultsModuleScoresType0", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                module_scores_type_0 = TestRunResultsModuleScoresType0.from_dict(data)

                return module_scores_type_0
            except:  # noqa: E722
                pass
            return cast(Union["TestRunResultsModuleScoresType0", None, Unset], data)

        module_scores = _parse_module_scores(d.pop("module_scores", UNSET))

        def _parse_module_group_scores(data: object) -> Union["TestRunResultsModuleGroupScoresType0", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                module_group_scores_type_0 = TestRunResultsModuleGroupScoresType0.from_dict(data)

                return module_group_scores_type_0
            except:  # noqa: E722
                pass
            return cast(Union["TestRunResultsModuleGroupScoresType0", None, Unset], data)

        module_group_scores = _parse_module_group_scores(d.pop("module_group_scores", UNSET))

        test_run_results = cls(
            id=id,
            created_at=created_at,
            updated_at=updated_at,
            name=name,
            test_id=test_id,
            test_name=test_name,
            agent_target_id=agent_target_id,
            agent_target_name=agent_target_name,
            status=status,
            label=label,
            test_label=test_label,
            agent_target_label=agent_target_label,
            started_at=started_at,
            completed_at=completed_at,
            result=result,
            result_score=result_score,
            result_score_change=result_score_change,
            links=links,
            conversation_scores=conversation_scores,
            metric_scores=metric_scores,
            module_scores=module_scores,
            module_group_scores=module_group_scores,
        )

        test_run_results.additional_properties = d
        return test_run_results

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="TestRunConversationScores")


@_attrs_define
class TestRunConversationScores:
    """
    Attributes:
        total_conversations (int):
        result_passed (int):
        result_failed (int):
        status_completed (int):
        status_error (int):
        status_running (int):
        status_pending (int):
        status_paused (int):
        status_canceled (int):
    """

    total_conversations: int
    result_passed: int
    result_failed: int
    status_completed: int
    status_error: int
    status_running: int
    status_pending: int
    status_paused: int
    status_canceled: int
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        total_conversations = self.total_conversations

        result_passed = self.result_passed

        result_failed = self.result_failed

        status_completed = self.status_completed

        status_error = self.status_error

        status_running = self.status_running

        status_pending = self.status_pending

        status_paused = self.status_paused

        status_canceled = self.status_canceled

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "total_conversations": total_conversations,
                "result_passed": result_passed,
                "result_failed": result_failed,
                "status_completed": status_completed,
                "status_error": status_error,
                "status_running": status_running,
                "status_pending": status_pending,
                "status_paused": status_paused,
                "status_canceled": status_canceled,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        total_conversations = d.pop("total_conversations")

        result_passed = d.pop("result_passed")

        result_failed = d.pop("result_failed")

        status_completed = d.pop("status_completed")

        status_error = d.pop("status_error")

        status_running = d.pop("status_running")

        status_pending = d.pop("status_pending")

        status_paused = d.pop("status_paused")

        status_canceled = d.pop("status_canceled")

        test_run_conversation_scores = cls(
            total_conversations=total_conversations,
            result_passed=result_passed,
            result_failed=result_failed,
            status_completed=status_completed,
            status_error=status_error,
            status_running=status_running,
            status_pending=status_pending,
            status_paused=status_paused,
            status_canceled=status_canceled,
        )

        test_run_conversation_scores.additional_properties = d
        return test_run_conversation_scores

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.metric import Metric


T = TypeVar("T", bound="ModuleMetrics")


@_attrs_define
class ModuleMetrics:
    """
    Attributes:
        id (str):
        created_at (datetime.datetime):
        updated_at (datetime.datetime):
        name (str):
        label (str):
        group_name (str):
        group_label (str):
        description (Union[None, Unset, str]):
        metrics (Union[Unset, List['Metric']]):
    """

    id: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    name: str
    label: str
    group_name: str
    group_label: str
    description: Union[None, Unset, str] = UNSET
    metrics: Union[Unset, List["Metric"]] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id = self.id

        created_at = self.created_at.isoformat()

        updated_at = self.updated_at.isoformat()

        name = self.name

        label = self.label

        group_name = self.group_name

        group_label = self.group_label

        description: Union[None, Unset, str]
        if isinstance(self.description, Unset):
            description = UNSET
        else:
            description = self.description

        metrics: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.metrics, Unset):
            metrics = []
            for metrics_item_data in self.metrics:
                metrics_item = metrics_item_data.to_dict()
                metrics.append(metrics_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "created_at": created_at,
                "updated_at": updated_at,
                "name": name,
                "label": label,
                "group_name": group_name,
                "group_label": group_label,
            }
        )
        if description is not UNSET:
            field_dict["description"] = description
        if metrics is not UNSET:
            field_dict["metrics"] = metrics

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.metric import Metric

        d = src_dict.copy()
        id = d.pop("id")

        created_at = isoparse(d.pop("created_at"))

        updated_at = isoparse(d.pop("updated_at"))

        name = d.pop("name")

        label = d.pop("label")

        group_name = d.pop("group_name")

        group_label = d.pop("group_label")

        def _parse_description(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        description = _parse_description(d.pop("description", UNSET))

        metrics = []
        _metrics = d.pop("metrics", UNSET)
        for metrics_item_data in _metrics or []:
            metrics_item = Metric.from_dict(metrics_item_data)

            metrics.append(metrics_item)

        module_metrics = cls(
            id=id,
            created_at=created_at,
            updated_at=updated_at,
            name=name,
            label=label,
            group_name=group_name,
            group_label=group_label,
            description=description,
            metrics=metrics,
        )

        module_metrics.additional_properties = d
        return module_metrics

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="Criteria")


@_attrs_define
class Criteria:
    """
    Attributes:
        id (str):
        scope_type (str):
        scope_id (str):
        criteria (str):
        criteria_label (str):
        criteria_description (Union[None, Unset, str]):
    """

    id: str
    scope_type: str
    scope_id: str
    criteria: str
    criteria_label: str
    criteria_description: Union[None, Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id = self.id

        scope_type = self.scope_type

        scope_id = self.scope_id

        criteria = self.criteria

        criteria_label = self.criteria_label

        criteria_description: Union[None, Unset, str]
        if isinstance(self.criteria_description, Unset):
            criteria_description = UNSET
        else:
            criteria_description = self.criteria_description

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "scope_type": scope_type,
                "scope_id": scope_id,
                "criteria": criteria,
                "criteria_label": criteria_label,
            }
        )
        if criteria_description is not UNSET:
            field_dict["criteria_description"] = criteria_description

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        id = d.pop("id")

        scope_type = d.pop("scope_type")

        scope_id = d.pop("scope_id")

        criteria = d.pop("criteria")

        criteria_label = d.pop("criteria_label")

        def _parse_criteria_description(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        criteria_description = _parse_criteria_description(d.pop("criteria_description", UNSET))

        criteria = cls(
            id=id,
            scope_type=scope_type,
            scope_id=scope_id,
            criteria=criteria,
            criteria_label=criteria_label,
            criteria_description=criteria_description,
        )

        criteria.additional_properties = d
        return criteria

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import List, Optional, Dict
import json
from pydantic import BaseModel
from datetime import datetime
from maihem.schemas.agents import AgentType
from maihem.api_client.maihem_client.models.conversation_nested import (
    ConversationNested,
)
from enum import Enum


class TestStatusEnum(str, Enum):
    PAUSED = "paused"
    PENDING = "pending"
    RUNNING = "running"
    COMPLETED = "completed"
    ERROR = "error"
    CANCELED = "canceled"


class TestResultEnum(str, Enum):
    PASSED = "passed"
    FAILED = "failed"


class APISchemaLinks(BaseModel):
    test_conversations: Optional[str] = None
    test_result: Optional[str] = None
    test_result_conversations: Optional[str] = None

    class Config:
        exclude_none = True


class Test(BaseModel):
    id: str
    created_at: datetime
    updated_at: datetime
    name: str
    agent_target_id: str
    label: Optional[str] = None
    initiating_agent: AgentType = AgentType.MAIHEM
    conversation_turns_max: Optional[int] = None
    agent_maihem_behavior_prompt: Optional[str] = None
    agent_maihem_goal_prompt: Optional[str] = None
    agent_maihem_population_prompt: Optional[str] = None
    metrics_config: Dict[str, int] = {}


class TestRun(BaseModel):
    id: str
    created_at: datetime
    updated_at: datetime
    test_id: str
    started_at: Optional[datetime] = None
    completed_at: Optional[datetime] = None
    status: TestStatusEnum
    result: Optional[TestResultEnum] = None
    links: Optional[APISchemaLinks] = None

    def __str__(self):
        return json.dumps(self.model_dump(), default=str, indent=4)


class TestRunConversations(TestRun):
    conversation_ids: List[str] = []

    def __str__(self):
        return json.dumps(self.model_dump(), default=str, indent=4)


class ConversationScores(BaseModel):
    total_conversations: int
    total_score: Optional[float] = None
    result_passed: int
    result_failed: int
    status_completed: int
    status_error: int
    status_running: int
    status_pending: int
    status_paused: int


class TestRunResultMetricScore(ConversationScores):
    result: Optional[TestResultEnum] = None
    status: TestStatusEnum


class TestRunResultMetrics(TestRun):
    conversation_scores: ConversationScores
    metric_scores: Dict[str, TestRunResultMetricScore]


class TestRunResultConversations(TestRun):
    conversations: List[ConversationNested] = []

    def __str__(self):
        test_run_dict = self.model_dump()
        conversation_jsons = [
            conversation.to_dict() for conversation in self.conversations
        ]
        test_run_dict["conversations"] = conversation_jsons

        return json.dumps(test_run_dict, default=str, indent=4)

    def to_json(self):
        return self.__str__()

    def to_dict(self):
        test_run_dict = self.model_dump()
        conversation_jsons = [
            conversation.to_dict() for conversation in self.conversations
        ]
        test_run_dict["conversations"] = conversation_jsons

        return test_run_dict

    class Config:
        arbitrary_types_allowed = True


class ResultEvaluation(BaseModel):
    criteria: str
    is_failure: bool
    explanation: Optional[str] = None

    def __str__(self):
        return json.dumps(self.model_dump(), default=str, indent=4)


class ResultConversationMessage(BaseModel):
    role: str
    content: str
    failed_evaluations: Optional[List[ResultEvaluation]] = None

    def __str__(self):
        return json.dumps(self.model_dump(), default=str, indent=4)


class ResultConversation(BaseModel):
    messages: List[ResultConversationMessage]
    failed_evaluations: Optional[List[ResultEvaluation]] = None

    def __str__(self):
        return json.dumps(self.model_dump(), default=str, indent=4)


class ResultTestRun(BaseModel):
    result: Optional[TestResultEnum] = None
    score: Optional[float] = None
    conversations: Optional[List[ResultConversation]] = None
    id: Optional[str] = None

    def __str__(self):
        return json.dumps(self.model_dump(), default=str, indent=4)

    def __init__(self, test_run_api: TestRunResultConversations):
        super().__init__()

        conversations = []
        for conv in test_run_api.conversations:
            messages = []
            for turn in conv.conversation_turns:
                for msg in turn.conversation_messages:
                    evaluations_msg = []
                    for evl in msg.evaluations:
                        is_failure = evl.result.value == "failed"
                        if is_failure:
                            evaluation_msg = ResultEvaluation(
                                criteria=evl.criteria,
                                is_failure=is_failure,
                                explanation=evl.explanation,
                            )
                            evaluations_msg.append(evaluation_msg)

                    message = ResultConversationMessage(
                        role=msg.agent_type.value,
                        content=msg.content,
                        failed_evaluations=evaluations_msg,
                    )
                    messages.append(message)

            evaluations = []
            for evl in conv.evaluations:
                is_failure = evl.result.value == "failed"
                if is_failure:
                    evaluation = ResultEvaluation(
                        criteria=evl.criteria,
                        is_failure=is_failure,
                        explanation=evl.explanation,
                    )
                    evaluations.append(evaluation)

            conversation = ResultConversation(
                messages=messages, failed_evaluations=evaluations
            )
            conversations.append(conversation)

        self.id = test_run_api.id
        self.result = test_run_api.result
        self.score = test_run_api.result_score
        self.conversations = conversations


# class SimulatedConversation:
#     """
#     Class with the response messages and evaluation from a simulated conversation

#     messages: List[Dict[str, str]] - List of messages from the conversation
#     evaluation: str - Explanation of the evaluation of the conversation
#     """

#     def __init__(self, conversation: TestRunResultConversations, conv_num: int = 0):
#         self.conv_num = conv_num
#         try:
#             self.messages = self._convert_conv_to_message_list(conversation)
#         except Exception as e:
#             self.messages = None
#         try:
#             self.evaluation = (
#                 conversation.conversations[conv_num].evaluations[0].explanation
#             )
#         except Exception as e:
#             self.evaluation = None

#     def _convert_conv_to_message_list(self, conversation: TestRunResultConversations):
#         conv = conversation.conversations[self.conv_num]
#         message_list = []

#         for turn in conv.conversation_turns:
#             for message in turn.conversation_messages:
#                 message_list.append({message.agent_type.value: message.content})

#         return message_list



"""Contains all the data models used in inputs/outputs"""

from .admin_test_result_metric_review_create_request import AdminTestResultMetricReviewCreateRequest
from .admin_test_result_metric_review_edit_request import AdminTestResultMetricReviewEditRequest
from .agent_target import AgentTarget
from .agent_target_create_request import AgentTargetCreateRequest
from .agent_type import AgentType
from .conversation_nested import ConversationNested
from .conversation_nested_evaluation import ConversationNestedEvaluation
from .conversation_nested_message import ConversationNestedMessage
from .conversation_nested_sentence import ConversationNestedSentence
from .conversation_nested_token_cost_base import ConversationNestedTokenCostBase
from .conversation_nested_turn import ConversationNestedTurn
from .conversation_turn_create_request import ConversationTurnCreateRequest
from .conversation_turn_create_response import ConversationTurnCreateResponse
from .create_test_run_request import CreateTestRunRequest
from .criteria import Criteria
from .criteria_instance_base import CriteriaInstanceBase
from .error_codes import ErrorCodes
from .error_response import ErrorResponse
from .error_response_error import ErrorResponseError
from .http_validation_error import HTTPValidationError
from .idp_org_create_request import IDPOrgCreateRequest
from .idp_org_user_add_request import IDPOrgUserAddRequest
from .idp_user_create_request import IDPUserCreateRequest
from .idp_user_update_request import IDPUserUpdateRequest
from .links import Links
from .metric import Metric
from .module import Module
from .module_metrics import ModuleMetrics
from .org import Org
from .org_create_request import OrgCreateRequest
from .test import Test
from .test_create_request import TestCreateRequest
from .test_create_request_documents_type_0 import TestCreateRequestDocumentsType0
from .test_create_request_metrics_config import TestCreateRequestMetricsConfig
from .test_documents_type_0 import TestDocumentsType0
from .test_metrics_config import TestMetricsConfig
from .test_result_enum import TestResultEnum
from .test_result_metric_feedback import TestResultMetricFeedback
from .test_result_metric_feedback_create_request import TestResultMetricFeedbackCreateRequest
from .test_result_metric_feedback_create_request_feedback import TestResultMetricFeedbackCreateRequestFeedback
from .test_result_metric_feedback_feedback import TestResultMetricFeedbackFeedback
from .test_run import TestRun
from .test_run_conversation_i_ds import TestRunConversationIDs
from .test_run_conversation_scores import TestRunConversationScores
from .test_run_metric_scores import TestRunMetricScores
from .test_run_metric_scores_criteria_failures_type_0 import TestRunMetricScoresCriteriaFailuresType0
from .test_run_results import TestRunResults
from .test_run_results_conversations import TestRunResultsConversations
from .test_run_results_conversations_metric_scores_type_0 import TestRunResultsConversationsMetricScoresType0
from .test_run_results_conversations_module_group_scores_type_0 import TestRunResultsConversationsModuleGroupScoresType0
from .test_run_results_conversations_module_scores_type_0 import TestRunResultsConversationsModuleScoresType0
from .test_run_results_metric_scores_type_0 import TestRunResultsMetricScoresType0
from .test_run_results_module_group_scores_type_0 import TestRunResultsModuleGroupScoresType0
from .test_run_results_module_scores_type_0 import TestRunResultsModuleScoresType0
from .test_run_status_update_request import TestRunStatusUpdateRequest
from .test_status_enum import TestStatusEnum
from .user import User
from .user_profile import UserProfile
from .v_test_result_metric_review_audit import VTestResultMetricReviewAudit
from .v_test_result_metric_review_state import VTestResultMetricReviewState
from .validation_error import ValidationError

__all__ = (
    "AdminTestResultMetricReviewCreateRequest",
    "AdminTestResultMetricReviewEditRequest",
    "AgentTarget",
    "AgentTargetCreateRequest",
    "AgentType",
    "ConversationNested",
    "ConversationNestedEvaluation",
    "ConversationNestedMessage",
    "ConversationNestedSentence",
    "ConversationNestedTokenCostBase",
    "ConversationNestedTurn",
    "ConversationTurnCreateRequest",
    "ConversationTurnCreateResponse",
    "CreateTestRunRequest",
    "Criteria",
    "CriteriaInstanceBase",
    "ErrorCodes",
    "ErrorResponse",
    "ErrorResponseError",
    "HTTPValidationError",
    "IDPOrgCreateRequest",
    "IDPOrgUserAddRequest",
    "IDPUserCreateRequest",
    "IDPUserUpdateRequest",
    "Links",
    "Metric",
    "Module",
    "ModuleMetrics",
    "Org",
    "OrgCreateRequest",
    "Test",
    "TestCreateRequest",
    "TestCreateRequestDocumentsType0",
    "TestCreateRequestMetricsConfig",
    "TestDocumentsType0",
    "TestMetricsConfig",
    "TestResultEnum",
    "TestResultMetricFeedback",
    "TestResultMetricFeedbackCreateRequest",
    "TestResultMetricFeedbackCreateRequestFeedback",
    "TestResultMetricFeedbackFeedback",
    "TestRun",
    "TestRunConversationIDs",
    "TestRunConversationScores",
    "TestRunMetricScores",
    "TestRunMetricScoresCriteriaFailuresType0",
    "TestRunResults",
    "TestRunResultsConversations",
    "TestRunResultsConversationsMetricScoresType0",
    "TestRunResultsConversationsModuleGroupScoresType0",
    "TestRunResultsConversationsModuleScoresType0",
    "TestRunResultsMetricScoresType0",
    "TestRunResultsModuleGroupScoresType0",
    "TestRunResultsModuleScoresType0",
    "TestRunStatusUpdateRequest",
    "TestStatusEnum",
    "User",
    "UserProfile",
    "ValidationError",
    "VTestResultMetricReviewAudit",
    "VTestResultMetricReviewState",
)



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.criteria import Criteria


T = TypeVar("T", bound="Metric")


@_attrs_define
class Metric:
    """
    Attributes:
        id (str):
        metric_group (str):
        metric_group_label (str):
        metric_sub_group (str):
        metric_sub_group_label (str):
        metric (str):
        metric_label (str):
        metric_description (Union[None, Unset, str]):
        criteria (Union[Unset, List['Criteria']]):
    """

    id: str
    metric_group: str
    metric_group_label: str
    metric_sub_group: str
    metric_sub_group_label: str
    metric: str
    metric_label: str
    metric_description: Union[None, Unset, str] = UNSET
    criteria: Union[Unset, List["Criteria"]] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id = self.id

        metric_group = self.metric_group

        metric_group_label = self.metric_group_label

        metric_sub_group = self.metric_sub_group

        metric_sub_group_label = self.metric_sub_group_label

        metric = self.metric

        metric_label = self.metric_label

        metric_description: Union[None, Unset, str]
        if isinstance(self.metric_description, Unset):
            metric_description = UNSET
        else:
            metric_description = self.metric_description

        criteria: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.criteria, Unset):
            criteria = []
            for criteria_item_data in self.criteria:
                criteria_item = criteria_item_data.to_dict()
                criteria.append(criteria_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "metric_group": metric_group,
                "metric_group_label": metric_group_label,
                "metric_sub_group": metric_sub_group,
                "metric_sub_group_label": metric_sub_group_label,
                "metric": metric,
                "metric_label": metric_label,
            }
        )
        if metric_description is not UNSET:
            field_dict["metric_description"] = metric_description
        if criteria is not UNSET:
            field_dict["criteria"] = criteria

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.criteria import Criteria

        d = src_dict.copy()
        id = d.pop("id")

        metric_group = d.pop("metric_group")

        metric_group_label = d.pop("metric_group_label")

        metric_sub_group = d.pop("metric_sub_group")

        metric_sub_group_label = d.pop("metric_sub_group_label")

        metric = d.pop("metric")

        metric_label = d.pop("metric_label")

        def _parse_metric_description(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        metric_description = _parse_metric_description(d.pop("metric_description", UNSET))

        criteria = []
        _criteria = d.pop("criteria", UNSET)
        for criteria_item_data in _criteria or []:
            criteria_item = Criteria.from_dict(criteria_item_data)

            criteria.append(criteria_item)

        metric = cls(
            id=id,
            metric_group=metric_group,
            metric_group_label=metric_group_label,
            metric_sub_group=metric_sub_group,
            metric_sub_group_label=metric_sub_group_label,
            metric=metric,
            metric_label=metric_label,
            metric_description=metric_description,
            criteria=criteria,
        )

        metric.additional_properties = d
        return metric

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Dict, List, Optional, Callable
import json

from maihem.api_client.maihem_client.client import Client as MaihemHTTPClient
from maihem.api_client.maihem_client.types import Response
from maihem.api_client.maihem_client.models.agent_target_create_request import (
    AgentTargetCreateRequest,
)
from maihem.api_client.maihem_client.models.agent_target import (
    AgentTarget,
)
from maihem.api_client.maihem_client.models.test_create_request import (
    TestCreateRequest,
)
from maihem.api_client.maihem_client.models.test import (
    Test,
)
from maihem.api_client.maihem_client.models.test_run import (
    TestRun,
)
from maihem.api_client.maihem_client.models.test_run_results_conversations import (
    TestRunResultsConversations,
)
from maihem.api_client.maihem_client.api.tests import tests_create_test
from maihem.api_client.maihem_client.api.tests import (
    tests_create_test_run,
    tests_get_tests,
    tests_get_test_test_runs,
)
from maihem.api_client.maihem_client.models.conversation_nested import (
    ConversationNested,
)
from maihem.api_client.maihem_client.models.conversation_turn_create_request import (
    ConversationTurnCreateRequest,
)
from maihem.api_client.maihem_client.models.conversation_turn_create_response import (
    ConversationTurnCreateResponse,
)
from maihem.api_client.maihem_client.api.test_runs import (
    test_runs_get_test_run,
    test_runs_get_test_run_conversations,
    test_runs_create_conversation_turn,
    test_runs_update_test_run_status,
    test_runs_get_test_run_result_with_conversations,
)
from maihem.api_client.maihem_client.models.create_test_run_request import (
    CreateTestRunRequest,
)
from maihem.api_client.maihem_client.models.test_run_status_update_request import (
    TestRunStatusUpdateRequest,
)
from maihem.api_client.maihem_client.api.conversations import (
    conversations_get_conversation,
)
from maihem.api_client.maihem_client.api.whoami import whoami_who_am_i
from maihem.api_client.maihem_client.api.agents import agents_create_agent_target
from maihem.api_client.maihem_client.api.agents import agents_get_agent_targets

from maihem.errors import handle_http_errors, ErrorResponse
from maihem.schemas.tests import TestStatusEnum
from maihem.logger import get_logger


class MaihemHTTPClientBase:
    def __init__(self, base_url: str, token: str):
        self.base_url = base_url
        self.token = token
        self._logger = get_logger()


class MaihemHTTPClientSync(MaihemHTTPClientBase):
    def whoami(self) -> Dict[str, str]:
        with MaihemHTTPClient(base_url=self.base_url) as client:
            response: Response = self._retry(whoami_who_am_i.sync_detailed)(
                client=client, x_api_key=self.token
            )

        if response.status_code != 200:
            handle_http_errors(error_resp=response.parsed)
        return response.parsed

    def _retry(self, function: Callable) -> Callable:
        def wrapper(*args, **kwargs):
            retries = 3
            for r in range(1, retries + 1):
                try:
                    return function(*args, **kwargs)
                except Exception as e:
                    if r == retries:
                        self._logger.error(
                            f"An error occurred in {function.__name__} after {retries} retries. Error: {e}"
                        )
                        raise e

            return None

        return wrapper

    def create_agent_target(self, req: AgentTargetCreateRequest) -> AgentTarget:
        with MaihemHTTPClient(base_url=self.base_url) as client:
            response: Response[AgentTarget] = self._retry(
                agents_create_agent_target.sync_detailed
            )(
                client=client,
                x_api_key=self.token,
                body=req,
            )

        if response.status_code != 201:
            error_dict = json.loads(response.content)
            handle_http_errors(error_resp=ErrorResponse.from_dict(error_dict))
        return response.parsed

    def get_agent_target_by_name(self, name: str) -> AgentTarget:

        with MaihemHTTPClient(base_url=self.base_url) as client:
            response: Response[List[AgentTarget]] = self._retry(
                agents_get_agent_targets.sync_detailed
            )(
                client=client,
                x_api_key=self.token,
                name=name,
            )

        if response.status_code != 200:
            error_dict = json.loads(response.content)
            handle_http_errors(error_resp=ErrorResponse.from_dict(error_dict))

        return response.parsed[0]

    def create_test(self, req: TestCreateRequest) -> Test:
        with MaihemHTTPClient(base_url=self.base_url) as client:
            response: Response[Test] = self._retry(tests_create_test.sync_detailed)(
                client=client,
                x_api_key=self.token,
                body=req,
            )

        if response.status_code != 201:
            error_dict = json.loads(response.content)
            handle_http_errors(error_resp=ErrorResponse.from_dict(error_dict))

        return response.parsed

    def get_test_by_name(self, name: str) -> Test:
        with MaihemHTTPClient(base_url=self.base_url) as client:
            response: Response[List[Test]] = self._retry(tests_get_tests.sync_detailed)(
                client=client,
                x_api_key=self.token,
                name=name,
            )

        if response.status_code != 200:
            error_dict = json.loads(response.content)
            handle_http_errors(error_resp=ErrorResponse.from_dict(error_dict))

        return response.parsed[0]

    def create_test_run(self, test_id: str, req: CreateTestRunRequest) -> TestRun:
        with MaihemHTTPClient(base_url=self.base_url) as client:
            response: Response[TestRun] = self._retry(
                tests_create_test_run.sync_detailed
            )(
                client=client,
                x_api_key=self.token,
                test_id=test_id,
                body=req,
            )

        if response.status_code != 201:
            error_dict = json.loads(response.content)
            handle_http_errors(error_resp=ErrorResponse.from_dict(error_dict))
        return response.parsed

    def update_test_run_status(
        self, test_run_id: str, status: TestStatusEnum
    ) -> TestRun:
        with MaihemHTTPClient(base_url=self.base_url) as client:
            response: Response[TestRun] = self._retry(
                test_runs_update_test_run_status.sync_detailed
            )(
                client=client,
                x_api_key=self.token,
                test_run_id=test_run_id,
                body=TestRunStatusUpdateRequest(status=status),
            )

        if response.status_code != 200:
            error_dict = json.loads(response.content)
            handle_http_errors(error_resp=ErrorResponse.from_dict(error_dict))
        return response.parsed

    def create_conversation_turn(
        self,
        test_run_id: str,
        conversation_id: str,
        target_agent_message: Optional[str] = None,
        contexts: Optional[List[str]] = None,
    ) -> ConversationTurnCreateResponse:
        with MaihemHTTPClient(base_url=self.base_url) as client:

            req: ConversationTurnCreateRequest = ConversationTurnCreateRequest(
                message=target_agent_message,
                contexts=contexts,
            )

            response: Response[ConversationTurnCreateResponse] = self._retry(
                test_runs_create_conversation_turn.sync_detailed
            )(
                client=client,
                x_api_key=self.token,
                test_run_id=test_run_id,
                conversation_id=conversation_id,
                body=req,
            )

        if response.status_code != 201:
            error_dict = json.loads(response.content)
            handle_http_errors(error_resp=ErrorResponse.from_dict(error_dict))
        return response.parsed

    def get_conversation(self, conversation_id: str) -> ConversationNested:
        with MaihemHTTPClient(base_url=self.base_url) as client:
            response: Response[ConversationNested] = self._retry(
                conversations_get_conversation.sync_detailed
            )(
                client=client,
                x_api_key=self.token,
                conversation_id=conversation_id,
            )

        if response.status_code != 200:
            error_dict = json.loads(response.content)
            handle_http_errors(error_resp=ErrorResponse.from_dict(error_dict))
        return response.parsed

    def get_test_run(self, test_run_id: str) -> TestRun:
        with MaihemHTTPClient(base_url=self.base_url) as client:
            response: Response[TestRun] = self._retry(
                test_runs_get_test_run.sync_detailed
            )(client=client, x_api_key=self.token, test_run_id=test_run_id)

        if response.status_code != 200:
            error_dict = json.loads(response.content)
            handle_http_errors(error_resp=ErrorResponse.from_dict(error_dict))
        return response.parsed

    def get_test_test_runs(self, test_id: str, test_run_name: str) -> List[TestRun]:
        with MaihemHTTPClient(base_url=self.base_url) as client:
            response: Response[List[TestRun]] = self._retry(
                tests_get_test_test_runs.sync_detailed
            )(
                client=client,
                x_api_key=self.token,
                test_id=test_id,
                name=test_run_name,
            )

        if response.status_code != 200:
            error_dict = json.loads(response.content)
            handle_http_errors(error_resp=ErrorResponse.from_dict(error_dict))
        return response.parsed

    def get_test_run_conversations(
        self, test_run_id: str
    ) -> TestRunResultsConversations:
        with MaihemHTTPClient(base_url=self.base_url) as client:
            response: Response[TestRunResultsConversations] = self._retry(
                test_runs_get_test_run_conversations.sync_detailed
            )(client=client, x_api_key=self.token, test_run_id=test_run_id)

        if response.status_code != 200:
            error_dict = json.loads(response.content)
            handle_http_errors(error_resp=ErrorResponse.from_dict(error_dict))
        return response.parsed

    # def get_test_run_result(self, name: str) -> TestRunResults:
    #     with MaihemHTTPClient(base_url=self.base_url) as client:
    #         response: Response[TestRunResults] = self._retry(
    #             test_runs_get_test_run_result.sync_detailed
    #         )(client=client, x_api_key=self.token, name=name)

    #     if response.status_code != 200:
    #         error_dict = json.loads(response.content)
    #         handle_http_errors(error_resp=ErrorResponse.from_dict(error_dict))
    #     return response.parsed

    def get_test_run_result(self, test_run_id: str) -> TestRunResultsConversations:
        with MaihemHTTPClient(base_url=self.base_url) as client:
            response: Response[TestRunResultsConversations] = self._retry(
                test_runs_get_test_run_result_with_conversations.sync_detailed
            )(client=client, x_api_key=self.token, test_run_id=test_run_id)

        if response.status_code != 200:
            error_dict = json.loads(response.content)
            handle_http_errors(error_resp=ErrorResponse.from_dict(error_dict))
        return response.parsed



from datetime import datetime
from pydantic import BaseModel

from enum import Enum
import os
from pydantic_extra_types.language_code import LanguageAlpha2
from typing import Callable, Optional, Tuple, List, Dict

import maihem.errors as errors
from maihem.logger import get_logger

logger = get_logger()


class AgentType(str, Enum):
    MAIHEM = "maihem"
    TARGET = "target"


class TargetAgent(BaseModel):
    id: str
    created_at: datetime
    updated_at: datetime
    name: str
    label: Optional[str] = None
    role: Optional[str] = None
    description: Optional[str] = None
    industry: Optional[str] = None
    language: Optional[LanguageAlpha2] = "en"

    _wrapper_function: Optional[Callable] = None
    document_paths: List[str] = []

    def set_wrapper_function(self, wrapper_function: Callable) -> None:
        is_test_success = self.test_wrapper_function(wrapper_function)

        if is_test_success:
            self._wrapper_function = wrapper_function
            logger.info("Wrapper function tested and added successfully")

    def test_wrapper_function(self, wrapper_function: Callable) -> bool:
        test_message = "Hi, it's the Maihem agent. Ready for testing?"
        logger.info("Testing wrapper function...")
        logger.info(f"Sending message to target agent: {test_message}")
        try:
            message, contexts = wrapper_function(str(datetime.now()), test_message, {})
            assert isinstance(message, str), "Response message must be a string"
            assert isinstance(contexts, list), "Contexts must be a list"
            for context in contexts:
                assert isinstance(
                    context, str
                ), "Each context in the list must be a string"

            return True
        except Exception as e:
            errors.raise_wrapper_function_error(
                f"Error testing target agent wrapper function: {e}"
            )

    def add_documents(self, documents: List[str]) -> None:
        for doc_path in documents:
            if not os.path.exists(doc_path):
                logger.warning(f"Document not found: {doc_path}")
                continue

            if not os.path.isfile(doc_path):
                logger.warning(f"Path is not a file: {doc_path}")
                continue

            file_size = os.path.getsize(doc_path)
            if file_size == 0:
                logger.warning(f"File is empty: {doc_path}")
                continue

            if file_size > 10 * 1024 * 1024:  # 10 MB limit
                logger.warning(f"File is too large (>10MB): {doc_path}")
                continue

            try:
                self.document_paths.append(doc_path)
                logger.info(f"Added document: {doc_path}")
            except Exception as e:
                logger.error(f"Error processing document {doc_path}: {str(e)}")

    def _send_message(
        self,
        conversation_id: str,
        message: Optional[str] = "",
        conversation_history: Dict = {},
    ) -> Tuple[str, List[str]]:
        if not self._wrapper_function:
            errors.raise_wrapper_function_error("Target agent wrapper function not set")
        for retry in range(3):
            try:
                response, contexts = self._wrapper_function(
                    conversation_id, message, conversation_history
                )
                return response, contexts
            except Exception as e:
                if retry < 2:
                    logger.warning(
                        f"Error sending message to target agent, retrying ({retry + 1}). Error: {str(e)}"
                    )
                else:
                    errors.raise_wrapper_function_error(
                        f"Error sending message to target agent after 3 retries. Error: {str(e)}"
                    )



from typing import Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.test_result_metric_feedback_create_request_feedback import TestResultMetricFeedbackCreateRequestFeedback
from ..types import UNSET, Unset

T = TypeVar("T", bound="TestResultMetricFeedbackCreateRequest")


@_attrs_define
class TestResultMetricFeedbackCreateRequest:
    """
    Attributes:
        entity_id (str):
        entity_type (str):
        test_result_metric_id (Union[None, Unset, str]):
        feedback (Union[Unset, TestResultMetricFeedbackCreateRequestFeedback]):  Default:
            TestResultMetricFeedbackCreateRequestFeedback.POSITIVE.
        comments (Union[None, Unset, str]):
    """

    entity_id: str
    entity_type: str
    test_result_metric_id: Union[None, Unset, str] = UNSET
    feedback: Union[Unset, TestResultMetricFeedbackCreateRequestFeedback] = (
        TestResultMetricFeedbackCreateRequestFeedback.POSITIVE
    )
    comments: Union[None, Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        entity_id = self.entity_id

        entity_type = self.entity_type

        test_result_metric_id: Union[None, Unset, str]
        if isinstance(self.test_result_metric_id, Unset):
            test_result_metric_id = UNSET
        else:
            test_result_metric_id = self.test_result_metric_id

        feedback: Union[Unset, str] = UNSET
        if not isinstance(self.feedback, Unset):
            feedback = self.feedback.value

        comments: Union[None, Unset, str]
        if isinstance(self.comments, Unset):
            comments = UNSET
        else:
            comments = self.comments

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "entity_id": entity_id,
                "entity_type": entity_type,
            }
        )
        if test_result_metric_id is not UNSET:
            field_dict["test_result_metric_id"] = test_result_metric_id
        if feedback is not UNSET:
            field_dict["feedback"] = feedback
        if comments is not UNSET:
            field_dict["comments"] = comments

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        entity_id = d.pop("entity_id")

        entity_type = d.pop("entity_type")

        def _parse_test_result_metric_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        test_result_metric_id = _parse_test_result_metric_id(d.pop("test_result_metric_id", UNSET))

        _feedback = d.pop("feedback", UNSET)
        feedback: Union[Unset, TestResultMetricFeedbackCreateRequestFeedback]
        if isinstance(_feedback, Unset):
            feedback = UNSET
        else:
            feedback = TestResultMetricFeedbackCreateRequestFeedback(_feedback)

        def _parse_comments(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        comments = _parse_comments(d.pop("comments", UNSET))

        test_result_metric_feedback_create_request = cls(
            entity_id=entity_id,
            entity_type=entity_type,
            test_result_metric_id=test_result_metric_id,
            feedback=feedback,
            comments=comments,
        )

        test_result_metric_feedback_create_request.additional_properties = d
        return test_result_metric_feedback_create_request

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.org import Org
    from ..models.user import User


T = TypeVar("T", bound="UserProfile")


@_attrs_define
class UserProfile:
    """
    Attributes:
        user (User):
        active_org_id (str):
        orgs (List['Org']):
    """

    user: "User"
    active_org_id: str
    orgs: List["Org"]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        user = self.user.to_dict()

        active_org_id = self.active_org_id

        orgs = []
        for orgs_item_data in self.orgs:
            orgs_item = orgs_item_data.to_dict()
            orgs.append(orgs_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "user": user,
                "active_org_id": active_org_id,
                "orgs": orgs,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.org import Org
        from ..models.user import User

        d = src_dict.copy()
        user = User.from_dict(d.pop("user"))

        active_org_id = d.pop("active_org_id")

        orgs = []
        _orgs = d.pop("orgs")
        for orgs_item_data in _orgs:
            orgs_item = Org.from_dict(orgs_item_data)

            orgs.append(orgs_item)

        user_profile = cls(
            user=user,
            active_org_id=active_org_id,
            orgs=orgs,
        )

        user_profile.additional_properties = d
        return user_profile

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class TestStatusEnum(str, Enum):
    CANCELED = "canceled"
    COMPLETED = "completed"
    ERROR = "error"
    PAUSED = "paused"
    PENDING = "pending"
    RUNNING = "running"

    def __str__(self) -> str:
        return str(self.value)



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.test_result_enum import TestResultEnum
from ..models.test_status_enum import TestStatusEnum
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.test_run_metric_scores_criteria_failures_type_0 import TestRunMetricScoresCriteriaFailuresType0


T = TypeVar("T", bound="TestRunMetricScores")


@_attrs_define
class TestRunMetricScores:
    """
    Attributes:
        total_conversations (int):
        result_passed (int):
        result_failed (int):
        status_completed (int):
        status_error (int):
        status_running (int):
        status_pending (int):
        status_paused (int):
        status_canceled (int):
        status (TestStatusEnum):
        total_score (Union[None, Unset, float]):
        total_score_change (Union[None, Unset, float]):
        result (Union[None, TestResultEnum, Unset]):
        summary (Union[None, Unset, str]):
        criteria_failures (Union['TestRunMetricScoresCriteriaFailuresType0', None, Unset]):
    """

    total_conversations: int
    result_passed: int
    result_failed: int
    status_completed: int
    status_error: int
    status_running: int
    status_pending: int
    status_paused: int
    status_canceled: int
    status: TestStatusEnum
    total_score: Union[None, Unset, float] = UNSET
    total_score_change: Union[None, Unset, float] = UNSET
    result: Union[None, TestResultEnum, Unset] = UNSET
    summary: Union[None, Unset, str] = UNSET
    criteria_failures: Union["TestRunMetricScoresCriteriaFailuresType0", None, Unset] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.test_run_metric_scores_criteria_failures_type_0 import TestRunMetricScoresCriteriaFailuresType0

        total_conversations = self.total_conversations

        result_passed = self.result_passed

        result_failed = self.result_failed

        status_completed = self.status_completed

        status_error = self.status_error

        status_running = self.status_running

        status_pending = self.status_pending

        status_paused = self.status_paused

        status_canceled = self.status_canceled

        status = self.status.value

        total_score: Union[None, Unset, float]
        if isinstance(self.total_score, Unset):
            total_score = UNSET
        else:
            total_score = self.total_score

        total_score_change: Union[None, Unset, float]
        if isinstance(self.total_score_change, Unset):
            total_score_change = UNSET
        else:
            total_score_change = self.total_score_change

        result: Union[None, Unset, str]
        if isinstance(self.result, Unset):
            result = UNSET
        elif isinstance(self.result, TestResultEnum):
            result = self.result.value
        else:
            result = self.result

        summary: Union[None, Unset, str]
        if isinstance(self.summary, Unset):
            summary = UNSET
        else:
            summary = self.summary

        criteria_failures: Union[Dict[str, Any], None, Unset]
        if isinstance(self.criteria_failures, Unset):
            criteria_failures = UNSET
        elif isinstance(self.criteria_failures, TestRunMetricScoresCriteriaFailuresType0):
            criteria_failures = self.criteria_failures.to_dict()
        else:
            criteria_failures = self.criteria_failures

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "total_conversations": total_conversations,
                "result_passed": result_passed,
                "result_failed": result_failed,
                "status_completed": status_completed,
                "status_error": status_error,
                "status_running": status_running,
                "status_pending": status_pending,
                "status_paused": status_paused,
                "status_canceled": status_canceled,
                "status": status,
            }
        )
        if total_score is not UNSET:
            field_dict["total_score"] = total_score
        if total_score_change is not UNSET:
            field_dict["total_score_change"] = total_score_change
        if result is not UNSET:
            field_dict["result"] = result
        if summary is not UNSET:
            field_dict["summary"] = summary
        if criteria_failures is not UNSET:
            field_dict["criteria_failures"] = criteria_failures

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.test_run_metric_scores_criteria_failures_type_0 import TestRunMetricScoresCriteriaFailuresType0

        d = src_dict.copy()
        total_conversations = d.pop("total_conversations")

        result_passed = d.pop("result_passed")

        result_failed = d.pop("result_failed")

        status_completed = d.pop("status_completed")

        status_error = d.pop("status_error")

        status_running = d.pop("status_running")

        status_pending = d.pop("status_pending")

        status_paused = d.pop("status_paused")

        status_canceled = d.pop("status_canceled")

        status = TestStatusEnum(d.pop("status"))

        def _parse_total_score(data: object) -> Union[None, Unset, float]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, float], data)

        total_score = _parse_total_score(d.pop("total_score", UNSET))

        def _parse_total_score_change(data: object) -> Union[None, Unset, float]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, float], data)

        total_score_change = _parse_total_score_change(d.pop("total_score_change", UNSET))

        def _parse_result(data: object) -> Union[None, TestResultEnum, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                result_type_0 = TestResultEnum(data)

                return result_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, TestResultEnum, Unset], data)

        result = _parse_result(d.pop("result", UNSET))

        def _parse_summary(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        summary = _parse_summary(d.pop("summary", UNSET))

        def _parse_criteria_failures(data: object) -> Union["TestRunMetricScoresCriteriaFailuresType0", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                criteria_failures_type_0 = TestRunMetricScoresCriteriaFailuresType0.from_dict(data)

                return criteria_failures_type_0
            except:  # noqa: E722
                pass
            return cast(Union["TestRunMetricScoresCriteriaFailuresType0", None, Unset], data)

        criteria_failures = _parse_criteria_failures(d.pop("criteria_failures", UNSET))

        test_run_metric_scores = cls(
            total_conversations=total_conversations,
            result_passed=result_passed,
            result_failed=result_failed,
            status_completed=status_completed,
            status_error=status_error,
            status_running=status_running,
            status_pending=status_pending,
            status_paused=status_paused,
            status_canceled=status_canceled,
            status=status,
            total_score=total_score,
            total_score_change=total_score_change,
            result=result,
            summary=summary,
            criteria_failures=criteria_failures,
        )

        test_run_metric_scores.additional_properties = d
        return test_run_metric_scores

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from typing import Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

T = TypeVar("T", bound="VTestResultMetricReviewAudit")


@_attrs_define
class VTestResultMetricReviewAudit:
    """
    Attributes:
        id (str):
        created_at (datetime.datetime):
        updated_at (datetime.datetime):
        test_result_metric_id (str):
        action (str):
        event (str):
        actioned_by (str):
        actioned_at (datetime.datetime):
        content (Union[None, str]):
        trm_entity_type (str):
        trm_entity_id (str):
        actioned_by_first_name (str):
        actioned_by_last_name (str):
        actioned_by_image_url (Union[None, str]):
        org_id (str):
        test_run_id (str):
        metric_id (str):
    """

    id: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    test_result_metric_id: str
    action: str
    event: str
    actioned_by: str
    actioned_at: datetime.datetime
    content: Union[None, str]
    trm_entity_type: str
    trm_entity_id: str
    actioned_by_first_name: str
    actioned_by_last_name: str
    actioned_by_image_url: Union[None, str]
    org_id: str
    test_run_id: str
    metric_id: str
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id = self.id

        created_at = self.created_at.isoformat()

        updated_at = self.updated_at.isoformat()

        test_result_metric_id = self.test_result_metric_id

        action = self.action

        event = self.event

        actioned_by = self.actioned_by

        actioned_at = self.actioned_at.isoformat()

        content: Union[None, str]
        content = self.content

        trm_entity_type = self.trm_entity_type

        trm_entity_id = self.trm_entity_id

        actioned_by_first_name = self.actioned_by_first_name

        actioned_by_last_name = self.actioned_by_last_name

        actioned_by_image_url: Union[None, str]
        actioned_by_image_url = self.actioned_by_image_url

        org_id = self.org_id

        test_run_id = self.test_run_id

        metric_id = self.metric_id

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "created_at": created_at,
                "updated_at": updated_at,
                "test_result_metric_id": test_result_metric_id,
                "action": action,
                "event": event,
                "actioned_by": actioned_by,
                "actioned_at": actioned_at,
                "content": content,
                "trm_entity_type": trm_entity_type,
                "trm_entity_id": trm_entity_id,
                "actioned_by_first_name": actioned_by_first_name,
                "actioned_by_last_name": actioned_by_last_name,
                "actioned_by_image_url": actioned_by_image_url,
                "org_id": org_id,
                "test_run_id": test_run_id,
                "metric_id": metric_id,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        id = d.pop("id")

        created_at = isoparse(d.pop("created_at"))

        updated_at = isoparse(d.pop("updated_at"))

        test_result_metric_id = d.pop("test_result_metric_id")

        action = d.pop("action")

        event = d.pop("event")

        actioned_by = d.pop("actioned_by")

        actioned_at = isoparse(d.pop("actioned_at"))

        def _parse_content(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        content = _parse_content(d.pop("content"))

        trm_entity_type = d.pop("trm_entity_type")

        trm_entity_id = d.pop("trm_entity_id")

        actioned_by_first_name = d.pop("actioned_by_first_name")

        actioned_by_last_name = d.pop("actioned_by_last_name")

        def _parse_actioned_by_image_url(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        actioned_by_image_url = _parse_actioned_by_image_url(d.pop("actioned_by_image_url"))

        org_id = d.pop("org_id")

        test_run_id = d.pop("test_run_id")

        metric_id = d.pop("metric_id")

        v_test_result_metric_review_audit = cls(
            id=id,
            created_at=created_at,
            updated_at=updated_at,
            test_result_metric_id=test_result_metric_id,
            action=action,
            event=event,
            actioned_by=actioned_by,
            actioned_at=actioned_at,
            content=content,
            trm_entity_type=trm_entity_type,
            trm_entity_id=trm_entity_id,
            actioned_by_first_name=actioned_by_first_name,
            actioned_by_last_name=actioned_by_last_name,
            actioned_by_image_url=actioned_by_image_url,
            org_id=org_id,
            test_run_id=test_run_id,
            metric_id=metric_id,
        )

        v_test_result_metric_review_audit.additional_properties = d
        return v_test_result_metric_review_audit

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from concurrent.futures import ThreadPoolExecutor, as_completed
import os
from typing import Dict, Literal, Optional, List, Callable
from pydantic import ValidationError
import random
from tqdm import tqdm
from yaspin import yaspin
from yaspin.spinners import Spinners

from maihem.modules_map import map_module_list_to_metrics
from maihem.schemas.agents import TargetAgent, AgentType
from maihem.schemas.tests import (
    Test,
    TestRun,
    TestRunConversations,
    ResultTestRun,
)
from maihem.schemas.conversations import ConversationTurnCreateResponse
from maihem.api_client.maihem_client.models.agent_target_create_request import (
    AgentTargetCreateRequest,
)
from maihem.api_client.maihem_client.models.test_create_request import (
    TestCreateRequest,
)
from maihem.api_client.maihem_client.models.test_create_request_metrics_config import (
    TestCreateRequestMetricsConfig,
)
from maihem.api_client.maihem_client.models.create_test_run_request import (
    CreateTestRunRequest,
)
from maihem.api_client.maihem_client.models.conversation_nested import (
    ConversationNested,
)
from maihem.api_client.maihem_client.models.conversation_nested_message import (
    ConversationNestedMessage,
)
import maihem.errors as errors
from maihem.api import MaihemHTTPClientSync
from maihem.schemas.tests import TestStatusEnum
from maihem.logger import get_logger
from maihem.utils.documents import TextSplitter, extract_text, parse_documents


class Client:
    _base_url: str = "https://api.maihem.ai"
    _base_url_ui: str = "https://cause.maihem.ai"
    _api_key: str = None

    _staging_url: str = "https://api.staging.maihem.ai"
    _staging_url_ui: str = "https://cause.staging.maihem.ai"

    _local_url: str = "http://localhost:8000"
    _local_url_ui: str = "http://localhost:3000"

    def create_target_agent(
        self,
        name: str,
        role: str,
        description: str,
        label: str,
        language: str,
    ) -> TargetAgent:
        pass

    def get_target_agent(self, name: str) -> TargetAgent:
        # Add implementation here
        raise NotImplementedError("Method not implemented")

    def create_test(
        self,
        name: str,
        target_agent_name: str,
        initiating_agent: Literal["maihem", "target"],
        label: Optional[str] = None,
        modules: str = None,
        metrics_config: Dict = None,
        maihem_agent_behavior_prompt: str = None,
        maihem_agent_goal_prompt: str = None,
        maihem_agent_population_prompt: str = None,
        conversation_turns_max: int = 10,
        number_conversations: int = 12,
    ) -> Test:
        raise NotImplementedError("Method not implemented")

    def run_test(
        self,
        name: str,
        test_name: str,
        wrapper_function: Callable,
        label: Optional[str] = None,
        concurrent_conversations: int = 10,
    ) -> TestRun:
        raise NotImplementedError("Method not implemented")

    def get_test_run_result(self, test_name: str, test_run_name: str) -> TestRun:
        raise NotImplementedError("Method not implemented")


class Maihem(Client):
    _maihem_api_client = MaihemHTTPClientSync

    def __init__(
        self,
        env: Optional[Literal["production", "staging", "local"]] = "production",
        api_key: Optional[str] = None,
    ) -> None:
        self._logger = get_logger()

        if env == "production":
            self._api_key = api_key or os.getenv("MAIHEM_API_KEY")
            if not self._api_key:
                raise errors.raise_request_validation_error("API key is missing")
            self._maihem_api_client = MaihemHTTPClientSync(
                self._base_url, self._api_key
            )
        elif env == "staging":
            self._api_key = api_key or os.getenv("MAIHEM_API_KEY_STAGING")
            if not self._api_key:
                raise errors.raise_request_validation_error("Staging API key missing")
            self._override_base_url(self._staging_url)
            self._override_base_url_ui(self._staging_url_ui)
        elif env == "local":
            self._api_key = api_key or os.getenv("MAIHEM_API_KEY_LOCAL")
            if not self._api_key:
                raise errors.raise_request_validation_error("Local API key missing")
            self._override_base_url(self._local_url)
            self._override_base_url_ui(self._local_url_ui)

    def _override_base_url(self, base_url: str) -> None:
        self._base_url = base_url
        self._maihem_api_client = MaihemHTTPClientSync(self._base_url, self._api_key)

    def _override_base_url_ui(self, base_url_ui: str) -> None:
        self._base_url_ui = base_url_ui

    def create_target_agent(
        self,
        name: str,
        role: str,
        description: str,
        label: Optional[str] = None,
        language: Optional[str] = "en",
    ) -> TargetAgent:
        logger = get_logger()
        logger.info(f"Creating target agent '{name}'...")
        resp = None
        try:
            resp = self._maihem_api_client.create_agent_target(
                req=AgentTargetCreateRequest(
                    name=name,
                    role=role,
                    label=label,
                    description=description,
                    language=language,
                )
            )
        except errors.ErrorBase as e:
            errors.handle_base_error(e)

        agent_target = None

        try:
            agent_target = TargetAgent.model_validate(resp.to_dict())
        except ValidationError as e:
            errors.handle_schema_validation_error(e)

        logger.info(f"Successfully created target agent '{name}'")
        return agent_target

    def get_target_agent(self, name: str) -> TargetAgent:
        resp = None
        try:
            resp = self._maihem_api_client.get_agent_target_by_name(name=name)
        except errors.ErrorBase as e:
            errors.handle_base_error(e)

        agent_target = None

        try:
            agent_target = TargetAgent.model_validate(resp.to_dict())
        except ValidationError as e:
            errors.handle_schema_validation_error(e)

        return agent_target

    def create_test(
        self,
        name: str,
        target_agent_name: str,
        initiating_agent: Optional[AgentType] = AgentType.MAIHEM,
        label: Optional[str] = None,
        modules: Optional[List[str]] = None,
        metrics_config: Optional[Dict] = None,
        maihem_behavior_prompt: Optional[str] = None,
        maihem_goal_prompt: Optional[str] = None,
        maihem_population_prompt: Optional[str] = None,
        conversation_turns_max: Optional[int] = 4,
        number_conversations: Optional[int] = 10,
        documents_path: Optional[str] = None,
    ) -> Test:
        logger = get_logger()
        logger.info(f"Creating test '{name}'...")

        # Input validation using pattern matching
        match (modules, metrics_config):
            case (None, None):
                raise ValueError("Either modules or metrics_config must be provided")
            case (list(), dict()):
                raise ValueError("Cannot provide both modules and metrics_config")
            case (list(), None):
                if not modules:
                    raise ValueError("Modules list must not be empty")
                metrics_config = map_module_list_to_metrics(
                    modules, number_conversations
                )
            case (None, dict()):
                if not metrics_config:
                    raise ValueError("Metrics config must not be empty")
                if not all(
                    isinstance(v, int) and v > 0 for v in metrics_config.values()
                ):
                    raise ValueError("Metrics config values must be positive integers")
            case _:
                raise ValueError("Invalid configuration for modules or metrics_config")

        # Convert string initiating_agent to enum if needed
        if isinstance(initiating_agent, str):
            try:
                initiating_agent = AgentType[initiating_agent.upper()]
            except KeyError:
                raise errors.raise_request_validation_error(
                    f"Invalid agent type: {initiating_agent}"
                )

        try:
            target_agent = self._maihem_api_client.get_agent_target_by_name(
                name=target_agent_name
            )

            if not target_agent:
                raise errors.raise_not_found_error(
                    f"Target agent '{target_agent_name}' not found"
                )

            if documents_path:
                documents = parse_documents(documents_path)
            else:
                documents = None

            with yaspin(
                Spinners.arc,
                text="Creating Test, this might take a minute...",
            ) as _:
                metrics_config_req = TestCreateRequestMetricsConfig.from_dict(
                    metrics_config
                )
                resp = self._maihem_api_client.create_test(
                    req=TestCreateRequest(
                        name=name,
                        label=label,
                        initiating_agent=initiating_agent,
                        conversation_turns_max=conversation_turns_max,
                        agent_target_id=target_agent.id,
                        agent_maihem_behavior_prompt=maihem_behavior_prompt,
                        agent_maihem_goal_prompt=maihem_goal_prompt,
                        agent_maihem_population_prompt=maihem_population_prompt,
                        metrics_config=metrics_config_req,
                        documents=documents if documents else None,
                    )
                )
        except errors.ErrorBase as e:
            errors.handle_base_error(e)

        test = None

        try:
            test = Test.model_validate(resp.to_dict())
        except ValidationError as e:
            errors.handle_schema_validation_error(e)

        logger.info(f"Successfully created test '{name}'")
        return test

    def get_test(self, name: str) -> Test:
        resp = None
        try:
            resp = self._maihem_api_client.get_test_by_name(name=name)
        except errors.ErrorBase as e:
            errors.handle_base_error(e)

        test = None

        try:
            test = Test.model_validate(resp.to_dict())
        except ValidationError as e:
            errors.handle_schema_validation_error(e)

        return test

    def run_test(
        self,
        name: str,
        test_name: str,
        wrapper_function: Callable,
        label: Optional[str] = None,
        concurrent_conversations: int = 10,
    ) -> ResultTestRun:
        logger = get_logger()
        test_run = None
        try:
            with yaspin(
                Spinners.arc,
                text=f"Preparing test run '{test_name}...",
            ) as _:
                test = self._maihem_api_client.get_test_by_name(name=test_name)
                target_agent = self.get_target_agent(name=test.agent_target_name)
                target_agent.set_wrapper_function(wrapper_function=wrapper_function)
                resp = None

            try:
                with yaspin(
                    Spinners.arc,
                    text="Creating Test Run, this might take a minute...",
                ) as _:
                    resp = self._maihem_api_client.create_test_run(
                        test_id=test.id,
                        req=CreateTestRunRequest(name=name, label=label),
                    )
            except errors.ErrorBase as e:
                errors.handle_base_error(e)

            try:
                test_run = TestRun.model_validate(resp.to_dict())
            except ValidationError as e:
                errors.handle_schema_validation_error(e)

            test_run_conversations = self.get_test_run_conversations(test_run.id)
            conversation_ids = test_run_conversations.conversation_ids

            logger.info(f"Starting test run '{test.name}'")
            print("\n" + "-" * 50 + "\n")
            logger.info(
                f"Running test run with {len(conversation_ids)} conversations (up to {concurrent_conversations} concurrently)..."
            )
            logger.info(f"Test run ID: {test_run.id}")
            logger.info(
                f"Test run results (UI): {self._base_url_ui}/evaluate/test-runs/{test_run.id}"
            )

            print("\n" + "-" * 50 + "\n")

            conversation_history = {}

            with tqdm(
                len(conversation_ids),
                total=len(conversation_ids),
                unit="conversation",
                colour="green",
                desc=f"Test run ({test_run.id})",
                position=0,
            ) as progress:
                with ThreadPoolExecutor(
                    max_workers=concurrent_conversations
                ) as executor:
                    future_to_conversation_id = {
                        executor.submit(
                            self._run_conversation,
                            test_run.id,
                            conversation_id,
                            test,
                            target_agent,
                            progress_bar_position=i + 1,
                            conversation_history=conversation_history,
                        ): conversation_id
                        for i, conversation_id in enumerate(conversation_ids)
                    }

                    for future in as_completed(future_to_conversation_id):
                        conversation_id = future_to_conversation_id[future]
                        try:
                            future.result()
                        except errors.ErrorBase as e:
                            logger.error(
                                f"Error running conversation ({conversation_id}): {e.message}"
                            )
                            progress.colour = "red"
                        finally:
                            progress.update()

            print("\n" + "-" * 50 + "\n")
            logger.info(f"Test run '{name}' completed")

            url_results = f"https://cause.maihem.ai/evaluate/test-runs/{test_run.id}"
            logger.info(f"See test run results: {url_results}")

            return self.get_test_run_result(test_name=test_name, test_run_name=name)

        except KeyboardInterrupt:
            logger.info("Keyboard interrupt detected. Canceling test...")
            if test_run is not None and test_run.id is not None:
                self._maihem_api_client.update_test_run_status(
                    test_run_id=test_run.id, status=TestStatusEnum.CANCELED
                )
            logger.info("Test run canceled!")
        except errors.ErrorBase as e:
            errors.handle_base_error(e)
        except Exception as e:
            logger.error(f"Error: {e}. Ending test...")
            if test_run is not None and test_run.id is not None:
                self._maihem_api_client.update_test_run_status(
                    test_run_id=test_run.id, status=TestStatusEnum.ERROR
                )

    def get_test_run_conversations(self, test_run_id: str) -> TestRunConversations:
        resp = None

        try:
            resp = self._maihem_api_client.get_test_run_conversations(test_run_id)
        except errors.ErrorBase as e:
            errors.handle_base_error(e)

        test_run = None

        try:
            test_run = TestRunConversations.model_validate(resp.to_dict())
        except ValidationError as e:
            errors.handle_schema_validation_error(e)

        return test_run

    def get_test_run_result(self, test_name: str, test_run_name: str) -> ResultTestRun:
        try:
            test = self.get_test(test_name)
        except errors.ErrorBase as e:
            errors.handle_base_error(e)

        try:
            test_runs = self._maihem_api_client.get_test_test_runs(
                test_id=test.id, test_run_name=test_run_name
            )
            if test_runs is None or len(test_runs) == 0:
                errors.raise_not_found_error(
                    f"Test run '{test_run_name}' not found for test '{test_name}'"
                )
        except errors.ErrorBase as e:
            errors.handle_base_error(e)

        test_run = None

        try:
            test_run_api = self._maihem_api_client.get_test_run_result(
                test_run_id=test_runs[0].id
            )
        except errors.ErrorBase as e:
            errors.handle_base_error(e)

        try:
            test_run = ResultTestRun(test_run_api=test_run_api)
        except ValidationError as e:
            errors.handle_schema_validation_error(e)

        return test_run

    # def get_test_run_result_conversations(
    #     self, test_run_id: str
    # ) -> TestRunResultConversations:
    #     resp = None
    #     test_run = None

    #     try:
    #         resp = self._maihem_api_client.get_test_run_result_conversations(
    #             test_run_id
    #         )
    #     except errors.ErrorBase as e:
    #         errors.handle_base_error(e)

    #     try:
    #         resp_conversations = resp.conversations
    #         resp_dict = resp.to_dict()
    #         resp_dict["conversations"] = []

    #         conversation_nesteds: ConversationNested = []
    #         for conv in resp_conversations:
    #             conv_dict = conv.to_dict()
    #             try:
    #                 conv = ConversationNested.from_dict(conv_dict)
    #                 conversation_nesteds.append(conv)
    #             except ValidationError as e:
    #                 errors.handle_schema_validation_error(e)

    #         test_run = TestRunResultConversations.model_validate(resp_dict)
    #         test_run.conversations = conversation_nesteds
    #     except ValidationError as e:
    #         errors.handle_schema_validation_error(e)

    #     return test_run

    def get_conversation(self, conversation_id: str) -> ConversationNested:
        resp = None

        try:
            resp = self._maihem_api_client.get_conversation(conversation_id)
        except errors.ErrorBase as e:
            errors.handle_base_error(e)

        try:
            conversation = ConversationNested.from_dict(resp.to_dict())
            return conversation
        except ValidationError as e:
            errors.handle_schema_validation_error(e)

    def _run_conversation(
        self,
        test_run_id: str,
        conversation_id: str,
        test: Test,
        target_agent: TargetAgent,
        progress_bar_position: int,
        conversation_history: Dict,
    ) -> str:
        is_conversation_active = True
        previous_turn_id = None
        turn_cnt = 0

        progress = tqdm(
            total=test.conversation_turns_max,
            desc=f"Conversation ({conversation_id})",
            unit="turn",
            position=progress_bar_position,
            leave=False,
        )
        while is_conversation_active:
            turn_cnt += 1
            progress.update()
            turn_resp = self._run_conversation_turn(
                test_run_id=test_run_id,
                conversation_id=conversation_id,
                test=test,
                target_agent=target_agent,
                previous_turn_id=previous_turn_id,
                conversation_history=conversation_history,
            )

            if (
                turn_resp.conversation.status != TestStatusEnum.RUNNING
                or not turn_resp.turn_id
            ):
                is_conversation_active = False
                progress.total = turn_cnt
                progress.n = turn_cnt
                progress.refresh()
                progress.close()
                return conversation_id

            previous_turn_id = turn_resp.turn_id

        return conversation_id

    def _run_conversation_turn(
        self,
        test_run_id: str,
        conversation_id: str,
        test: Test,
        target_agent: TargetAgent,
        conversation_history: Dict,
        previous_turn_id: Optional[str] = None,
    ) -> ConversationTurnCreateResponse:
        agent_maihem_message = None

        conversation = self.get_conversation(conversation_id)

        document_key = None
        text = None

        logger = get_logger()

        # Document loading and chunking (for RAG)
        if target_agent.document_paths:
            max_attempts = 10
            attempts = 0
            while attempts < max_attempts:
                if not target_agent.document_paths:
                    break
                document_path = random.choice(list(target_agent.document_paths))
                document_key = os.path.basename(document_path)
                try:
                    document = extract_text(document_path)
                    if len(document) > 10000:
                        chunks = TextSplitter(
                            chunk_size=5000, chunk_overlap=200
                        ).split_text(document)
                        text = random.choice(chunks)
                    else:
                        text = document
                    if text.strip():
                        break
                except Exception as e:
                    logger.warning(
                        f"Error processing document {document_key}: {str(e)}"
                    )
                attempts += 1

            if attempts == max_attempts:
                logger.warning(
                    "Max attempts reached while trying to select a valid document chunk."
                )

        if (
            test.initiating_agent == AgentType.MAIHEM
            and len(conversation.conversation_turns) == 0
        ):

            turn_resp = self._generate_conversation_turn(
                test_run_id=test_run_id,
                conversation_id=conversation_id,
                target_agent_message=None,
                contexts=[],
            )

            agent_maihem_message = self._get_conversation_message_from_conversation(
                turn_id=turn_resp.turn_id,
                agent_type=AgentType.MAIHEM,
                conversation=turn_resp.conversation,
            )
        elif previous_turn_id is None and len(conversation.conversation_turns) > 0:
            return ConversationTurnCreateResponse(
                turn_id=None, conversation=conversation
            )
        elif previous_turn_id is not None and len(conversation.conversation_turns) > 0:
            agent_maihem_message = self._get_conversation_message_from_conversation(
                turn_id=previous_turn_id,
                agent_type=AgentType.MAIHEM,
                conversation=conversation,
            )

        try:
            agent_maihem_message = (
                agent_maihem_message.content if agent_maihem_message else None
            )
            target_agent_message, contexts = target_agent._send_message(
                conversation_id, agent_maihem_message, conversation_history
            )
        except Exception as e:
            errors.raise_wrapper_function_error(
                f"Error sending message to target agent: {e}"
            )

        if contexts != []:
            contexts_concat = "\n".join(contexts)
            if len(contexts_concat) > 20000:
                errors.raise_chat_function_error(
                    "Length of all contexts combined should not exceed 20,000 characters"
                )

        turn_resp = self._generate_conversation_turn(
            test_run_id=test_run_id,
            conversation_id=conversation_id,
            target_agent_message=target_agent_message,
            contexts=contexts,
        )

        return turn_resp

    def _generate_conversation_turn(
        self,
        test_run_id: str,
        conversation_id: str,
        target_agent_message: Optional[str] = None,
        contexts: Optional[List[str]] = None,
    ) -> ConversationTurnCreateResponse:
        try:
            resp = self._maihem_api_client.create_conversation_turn(
                test_run_id=test_run_id,
                conversation_id=conversation_id,
                target_agent_message=target_agent_message,
                contexts=contexts,
            )

            return ConversationTurnCreateResponse(
                turn_id=resp.turn_id,
                conversation=resp.conversation,
            )
        except ValidationError as e:
            errors.handle_schema_validation_error(e)
        except errors.ErrorBase as e:
            errors.handle_base_error(e)

    def _get_conversation_message_from_conversation(
        self,
        turn_id: str,
        agent_type: AgentType,
        conversation: ConversationNested,
    ) -> ConversationNestedMessage:
        for turn in conversation.conversation_turns:

            if turn.id == turn_id:
                for message in turn.conversation_messages:
                    if message.agent_type == agent_type:
                        return message

        errors.raise_not_found_error(f"Could not retrieve {agent_type} agent message")



import os
from openai import AzureOpenAI, AsyncAzureOpenAI, AsyncOpenAI


class Agent:

    def __init__(self, instruction_prompt: str, temperature: float = 0.6):
        self.instruction_prompt = instruction_prompt
        self.temperature = temperature
        self.messages = [{"role": "system", "content": instruction_prompt}]

    def chat(self, message: str) -> str:

        if message:
            self.messages.append({"role": "user", "content": message})
            response = self.__llm_call_azure()
        else:
            response = "Hi, how can I help you?"
        self.messages.append({"role": "assistant", "content": response})
        return response

    async def achat(self, message: str) -> str:
        self.messages.append({"role": "user", "content": message})
        response = await self.__async_llm_call_azure()
        self.messages.append({"role": "assistant", "content": response})
        return response

    def __llm_call_azure(self):
        """
        Call the Azure OpenAI API asynchronously with a single prompt.
        Returns a completion for the prompt.
        """
        client = AzureOpenAI(
            api_key=os.getenv("AZURE_OPENAI_API_KEY"),
            api_version="2024-05-01-preview",
            azure_endpoint=os.getenv("AZURE_OPENAI_ENDPOINT"),
        )

        response = client.chat.completions.create(
            model="maihem-gpt4o",  # deployment_name
            messages=self.messages,
            temperature=self.temperature,
        )

        response_msg = response.choices[0].message.content

        return response_msg

    async def __async_llm_call_azure(self):
        """
        Call the Azure OpenAI API asynchronously with a single prompt.
        Returns a completion for the prompt.
        """
        client = AsyncAzureOpenAI(
            api_key=os.getenv("AZURE_OPENAI_API_KEY"),
            api_version="2024-05-01-preview",
            azure_endpoint=os.getenv("AZURE_OPENAI_ENDPOINT"),
        )

        response = await client.chat.completions.create(
            model="maihem-gpt4o",  # deployment_name
            messages=self.messages,
            temperature=self.temperature,
        )

        response_msg = response.choices[0].message.content

        return response_msg
